/*            Copyright (C) 2000, 2001, 2002, 2003 Stijn van Dongen
 *
 * This file is part of MCL.  You can redistribute and/or modify MCL under the
 * terms of the GNU General Public License; either version 2 of the License or
 * (at your option) any later version.  You should have received a copy of the
 * GPL along with MCL, in the file COPYING.
*/

#include "array.h"
#include "alloc.h"
#include "types.h"
#include "err.h"
#include <string.h>

void  mcxSplice
(  void*           base1pptr
,  const void*     base2ptr
,  int             size      /*  size of base1 and base2 members          */
,  int         *pn_base1     /*  # base1 elements currently in use        */
,  int         *pN_base1     /*  # base1 elements for which is malloced   */
,  int           O_base1     /*  splice relative to this element          */
,  int           d_base1     /*  delete this number of elements           */
,  int           c_base2     /*  number of elements to copy               */
)  
   {  char **ppr1          =  (char **)   base1pptr
   ;  const char *ptr2     =  (const char *)   base2ptr

   ;  int   n_base1        =  *pn_base1
   ;  int   N_base1        =  *pN_base1
   ;  int   m_base1        =  n_base1 - d_base1 + c_base2

   ;  int   o_base1        =  O_base1 >= 0
                              ?  
                                 O_base1
                              :  
                                 n_base1 + O_base1 + 1

   ;  const char  *dieMsg  =  ""

   ;  if (0)
      {  die
      :  fprintf
         (  stderr
         ,  "[mcxSplice PBD] %s\n"
            "[mcxSplice] [n1, %d] [N1, %d] [o1, %d] [d1, %d] [c2, %d]\n"
         ,  dieMsg
         ,  n_base1, N_base1, O_base1, d_base1, c_base2
         )
      ;  mcxExit(1)
   ;  }

      if (   n_base1    <  0
         ||  N_base1    <  0
         ||  n_base1    >  N_base1
         ||  d_base1    <  0
         ||  c_base2    <  0
         )
      {  dieMsg      =  "integer arguments not consistent"
      ;  goto die
   ;  }

      if (o_base1 < 0 || o_base1 > n_base1)
      {  dieMsg      =  "computed splice offset not in bounds"
      ;  goto die
   ;  }

      if (*ppr1 == NULL && ptr2 == NULL)
      {  dieMsg      =  "source and destination both void"
      ;  goto die
   ;  }

      if (o_base1 + d_base1 > n_base1)
      {  dieMsg      =  "not that many elements to delete"
      ;  goto die
   ;  }

      if (m_base1 > N_base1)
         *ppr1       =  mcxRealloc(*ppr1, size*m_base1, RETURN_ON_FAIL)
      ,  *pN_base1   =  N_base1 = m_base1

   ;  if (m_base1 && *ppr1 == NULL)
         mcxMemDenied(stderr, "mcxSplice", "void", m_base1)
      ,  mcxExit(1)

   ;  if (o_base1 < n_base1)
      memmove
      (  *ppr1 + size*(o_base1 + c_base2)
      ,  *ppr1 + size*(o_base1 + d_base1)
      ,  size*(n_base1 - o_base1 - d_base1)
      )

   ;  if (c_base2)
      memcpy
      (  *ppr1 + size * (o_base1)
      ,  ptr2
      ,  size*(c_base2)
      )
   ;  *pn_base1      =  m_base1
;  }


int mcxDedup
(  void*                base
,  int                  nmemb
,  int                  size
,  int                  (*cmp)(const void *, const void *)
,  void                 (*merge)(void *, const void *)
)  
   {  int   k  =  0
   ;  int   l  =  0 
      
   ;  while (l < nmemb)
      {  
         if (k != l)
         memcpy(((char*)base) + k * size, ((char*)base) + l * size, size)

      ;  while
         (  ++l < nmemb
         && (  cmp
            ?  (!cmp(((char*)base) + k * size, ((char*)base) + l * size))
            :  (!memcmp(((char*)base) + k*size, ((char*)base) + l*size, size))
            )  
         )  
         {  if (merge)
            merge(((char*)base) + k * size, ((char*)base) + l * size)
      ;  }
         k++
   ;  }

      return k       
;  }


mcxstatus mcxResize
(  void*          mempp
,  int            size
,  int*           ct
,  int            newct
,  mcxOnFail      ON_FAIL
)
   {  char **pp   =  (char **) mempp
   ;  char* ptr   =  *pp
   ;  if (newct && !(ptr = mcxRealloc(ptr, size*newct, ON_FAIL)))
      {  *pp      =  NULL
      ;  *ct      =  0
      ;  return STATUS_FAIL
   ;  }
      *pp   =  ptr
   ;  *ct   =  newct
   ;  return STATUS_OK
;  }


