/*            Copyright (C) 2000, 2001, 2002, 2003 Stijn van Dongen
 *
 * This file is part of MCL.  You can redistribute and/or modify MCL under the
 * terms of the GNU General Public License; either version 2 of the License or
 * (at your option) any later version.  You should have received a copy of the
 * GPL along with MCL, in the file COPYING.
*/

#include <stdio.h>

#include "equate.h"
#include "err.h"
#include "checkbounds.h"


int checkBoundsUsage
(  char        type
,  void*       var
,  int         (*lftRlt) (const void*, const void*)
,  void*       lftBound
,  int         (*rgtRlt) (const void*, const void*)
,  void*       rgtBound
)  ;


int (*rltFunctions[8])(const void* f1, const void* f2) =  
{  intGt  ,  intGq  ,  fltGt,  fltGq
,  intLt  ,  intLq  ,  fltLt,  fltLq
}  ;


int checkBoundsPrintDigits       =  5;
char* flagCheckBoundsCaller      =  (char *) "mcl";


const char* rltSigns[8] =
{  "(",  "[",  "(",  "["
,  ")",  "]",  ")",  "]"
}  ;


/*    returns
 *    0  for matching bounds
 *    1  for non-matching bounds
*/

int checkBounds
(  char        type
,  void*       var
,  int         (*lftRlt) (const void*, const void*)
,  void*       lftBound
,  int         (*rgtRlt) (const void*, const void*)
,  void*       rgtBound
)
   {  int lftOk, rgtOk
   ;  if (checkBoundsUsage(type, var, lftRlt, lftBound, rgtRlt, rgtBound))
      {  mcxErr("checkBounds", "fatal")
      ;  mcxExit(1)
   ;  }
      lftOk =   !lftRlt || lftRlt(var, lftBound)
   ;  rgtOk =   !rgtRlt || rgtRlt(var, rgtBound)
   ;  return (lftOk && rgtOk) ? 0 : 1
;  }


/* todo: NULL lftbound argument & non-NULL lftRelate argument */
/* idem for rgt */

int checkBoundsUsage
(  char        type
,  void*       var
,  int         (*lftRlt) (const void*, const void*)
,  void*       lftBound
,  int         (*rgtRlt) (const void*, const void*)
,  void*       rgtBound
)
   {  int  i
   ;  const char* me  = "checkBoundsUsage"

   ;  if (type != 'f' && type != 'i')
      {  mcxErr(me, "pbd: unsupported checkbound type <%c>", type)
      ;  return 1
   ;  }

      if ((lftRlt && !lftBound)||(!lftRlt && lftBound))
      {  mcxErr(me, "pbd: abusive lftRlt lftBound combination")
      ;  return 1
   ;  }
      if ((rgtRlt && !rgtBound)||(!rgtRlt && rgtBound))
      {  mcxErr(me, "pbd: abusive rgtRlt rgtBound combination")
      ;  return 1
   ;  }

      if (lftRlt)
      {  for(i=0;i<4;i++) if (lftRlt == rltFunctions[i]) break
      ;  if (i == 4)
         {  mcxErr(me, "pbd: lftRlt should use gt or gq arg")
         ;  return 1
      ;  }
   ;  }

      if (rgtRlt)
      {  for(i=4;i<8;i++) if (rgtRlt == rltFunctions[i]) break
      ;  if (i==8)
         {  mcxErr(me, "pbd: rgtRlt should use lt or lq arg")
         ;  return 1
      ;  }
   ;  }
      return 0
;  }


mcxTing* checkBoundsRange
(  char        type
,  void*       var
,  int         (*lftRlt) (const void*, const void*)
,  void*       lftBound
,  int         (*rgtRlt) (const void*, const void*)
,  void*       rgtBound
)
   {  mcxTing*  textRange   =  mcxTingEnsure(NULL, 40)
   ;  char* lftToken       =  (char *) "<?"
   ;  char* rgtToken       =  (char *) "?>"
   ;  int i

   ;  if (lftRlt)
      {  for(i=0;i<4;i++)
            if (lftRlt == rltFunctions[i])
               break
      ;  if (i<4)
            lftToken = (char *) rltSigns[i]
   ;  }
      else
      {  lftToken = (char *) "("
   ;  }

      textRange->len += sprintf( textRange->str+textRange->len, "%s", lftToken )

   ;  if (lftBound)
      {  if (type == 'f')
         {  textRange->len +=
            sprintf
            (  textRange->str+textRange->len, "%.*f"
            ,  checkBoundsPrintDigits, *((float*)lftBound)
            )
      ;  }
         else if (type == 'i')  
         {  textRange->len +=
            sprintf
            (  textRange->str+textRange->len, "%d", *((int*)lftBound) )
      ;  }
   ;  }
      else
      {  textRange->len += sprintf(textRange->str+textRange->len, "%s", "<-")
   ;  }

      textRange->len += sprintf(textRange->str+textRange->len, "%s", ",")

   ;  if (rgtBound)
      {  if (type == 'f')
         {  textRange->len +=
            sprintf
            (  textRange->str+textRange->len, "%.*f"
            ,  checkBoundsPrintDigits, *((float*)rgtBound)
            )
           /* BUG fixme; this can overrun buffer */
      ;  }
         else if (type == 'i')
         {  textRange->len +=
            sprintf(textRange->str+textRange->len, "%d", *((int*)rgtBound))
      ;  }
   ;  }
      else
      {  textRange->len += sprintf(textRange->str+textRange->len, "%s", "->")
   ;  }

   ;  if (rgtRlt)
      {  for(i=4;i<8;i++)
            if (rgtRlt == rltFunctions[i])
               break
      ;  if (i<8)
            rgtToken = (char *) rltSigns[i]
   ;  }
      else
      rgtToken = (char *) ")"

   ;  textRange->len += sprintf(textRange->str+textRange->len, "%s", rgtToken)

   ;  return textRange
;  }


mcxbool flagCheckBounds
(  const char*    caller
,  const char*    flag
,  char           type
,  void*          var
,  int            (*lftRlt) (const void*, const void*)
,  void*          lftBound
,  int            (*rgtRlt) (const void*, const void*)
,  void*          rgtBound
)
   {  mcxTing* textRange

   ;  if (checkBounds(type, var, lftRlt, lftBound, rgtRlt, rgtBound))
      {  textRange =
         checkBoundsRange(type, var, lftRlt, lftBound, rgtRlt, rgtBound)
      ;  mcxErr
         (  caller
         ,  "%s argument to %s should be in range %s"
         ,  type == 'i' ? "integer" : type == 'f' ? "float" : "MICANS"
         ,  flag
         ,  textRange->str
         )
      ;  mcxTingFree(&textRange)
      ;  return FALSE
   ;  }
      return TRUE
;  }

