/*            Copyright (C) 2002, 2003 Stijn van Dongen
 *
 * This file is part of MCL.  You can redistribute and/or modify MCL under the
 * terms of the GNU General Public License; either version 2 of the License or
 * (at your option) any later version.  You should have received a copy of the
 * GPL along with MCL, in the file COPYING.
*/

#include <string.h>

#include "opt.h"
#include "types.h"
#include "err.h"

mcxHash* mcxOptHash
(  mcxOptAnchor*  opts
,  mcxHash*       hash
)
   {  mcxOptAnchor*  anch  =  opts ? opts+0 : NULL

   ;  hash  =  hash
               ?  hash
               :  mcxHashNew
                  (  8
                  ,  mcxStrHash
                  ,  (int (*)(const void*, const void*))strcmp
                  )

   ;  while (anch && anch->tag)
      {  mcxKV*   kv       =  mcxHashSearch(anch->tag, hash, MCX_DATUM_INSERT)
      ;  if (kv->val)
         mcxWarn
         ("mcxOptHash", "warning: option <%s> already present", anch->tag)
      ;  kv->val = anch
      ;  anch++
   ;  }
      return hash
;  }


void mcxOptHashFree
(  mcxHash**   hashpp
)
   {  mcxHashFree(hashpp, NULL, NULL)
;  }


mcxOptList* mcxOptParse__
(  mcxHash*       opthash
,  char**         argv
,  int            argc
,  int            prefix   /* skip these */
,  int            suffix   /* skip those too */
,  int*           n_elems_read
,  mcxstatus*     status
)
   {  char**   argp  =  argv+prefix
   ;  char**   argl  =  argv+argc-suffix     /* arg last */
   ;  mcxbool  do_exhaust = n_elems_read ? TRUE : FALSE
                                       /* fixme: very ugly internal iface*/

   ;  mcxOptList  boot
   ;  mcxOptList* tail = &boot

   ;  *status     =  MCX_OPT_OK
   ;  tail->next  =  NULL
   ;  tail->tag   =  NULL

   ;  if (do_exhaust)
      *n_elems_read = 0

   ;  while (argp < argl)
      {  char* arg            =  *argp
      ;  mcxKV* kv            =  mcxHashSearch(arg, opthash, MCX_DATUM_FIND)
      ;  mcxOptAnchor* anch   =  kv ? (mcxOptAnchor*) kv->val : NULL
      ;  mcxOptList* t        =  mcxAlloc(sizeof(mcxOptList), EXIT_ON_FAIL)

      ;  t->val = NULL

      ;  if (kv)
         {  if (do_exhaust)
            (*n_elems_read)++

         ;  if (anch->flags & MCX_OPT_HASARG) 
            {  argp++
            ;  if (argp >= argl)
               {  mcxErr("mcxOptParse", "option <%s> takes value", anch->tag)
               ;  *status  =  MCX_OPT_NOARG
               ;  return NULL    /* fixme: cleanup !!? */
                              /* Null return / status value interface issue */
            ;  }
               t->val = *argp    /* mq note: shallow copy */
            ;  if (do_exhaust)
               (*n_elems_read)++
         ;  }
         }
         else
         {  if (do_exhaust)
            {  mcxFree(t)
            ;  return boot.next
         ;  }
            else
            {  mcxErr("mxcOptParse", "unsupported option <%s>", arg)
            ;  *status  =  MCX_OPT_UNKNOWN
            ;  return NULL    /* fixme: cleanup !!? */
         ;  }
         }

         tail->next  =  t
      ;  t->tag      =  anch->tag
      ;  t->next     =  NULL
      ;  tail        =  t

      ;  argp++
   ;  }

      return boot.next
;  }


mcxOptList* mcxOptExhaust
(  mcxHash*       opthash
,  char**         argv
,  int            argc
,  int            prefix   /* skip these */
,  int*           n_elems_read
,  mcxstatus*     status
)
   {  return mcxOptParse__
      (opthash, argv, argc, prefix, 0, n_elems_read, status)
;  }


mcxOptList* mcxOptParse
(  mcxHash*       opthash
,  char**         argv
,  int            argc
,  int            prefix   /* skip these */
,  int            suffix   /* skip those too */
,  mcxstatus*     status
)
   {  return mcxOptParse__(opthash, argv, argc, prefix, suffix, NULL, status)
;  }


void mcxOptListFree
(  mcxOptList** listpp
)
   {  mcxOptList* node = *listpp
   ;  mcxOptList* next = node ? node->next : NULL

   ;  while (node)
      {  next = node->next
      ;  mcxFree(node)
      ;  node = next
   ;  }
   }


void mcxUsage
(  FILE* fp
,  const char*  caller
,  const char** lines
)
   {  int i =  0

   ;  while(lines[i])
      {  fprintf(fp, "%s\n", lines[i])
      ;  i++
   ;  }
      fprintf(fp, "[%s] Printed %d lines\n", caller, i+1)
;  }

